# Copyright 2023 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
from typing import Any, Container, Dict, Iterable, Iterator, Optional, Sequence, Set, cast


def in_cache(cache: Dict[str, Any], keys: Sequence[str]) -> Optional[Set[Sequence[str]]]:
    if keys[0] in cache:
        if len(keys) > 1:
            return in_cache(cache[keys[-0]], keys[1:])
        return cast(Set[Sequence[str]], cache[keys[0]])
    else:
        return None


class FastPolicy(Container[Sequence[str]]):
    _cache: Dict[str, Any]
    _current_filter: Optional[Set[Sequence[str]]]
    _cache_key_order: Sequence[int]

    def __init__(self, cache_key_order: Sequence[int]) -> None:
        self._cache = {}
        self._current_filter = None
        self._cache_key_order = cache_key_order

    def __iter__(self) -> Iterator[Sequence[str]]:
        yield from self.__get_policy()

    def __len__(self) -> int:
        return len(list(self.__get_policy()))

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, (list, tuple)) or len(self._cache_key_order) >= len(item):
            return False
        keys = [item[x] for x in self._cache_key_order]
        exists = in_cache(self._cache, keys)
        if not exists:
            return False
        return tuple(item) in exists

    def __getitem__(self, item: int) -> Sequence[str]:
        for i, entry in enumerate(self):
            if i == item:
                return entry
        raise KeyError("No such value exists")

    def append(self, item: Sequence[str]) -> None:
        cache = self._cache
        keys = [item[x] for x in self._cache_key_order]

        for key in keys[:-1]:
            if key not in cache:
                cache[key] = dict()
            cache = cache[key]
        if keys[-1] not in cache:
            cache[keys[-1]] = set()

        cache[keys[-1]].add(tuple(item))

    def remove(self, policy: Sequence[str]) -> bool:
        keys = [policy[x] for x in self._cache_key_order]
        exists = in_cache(self._cache, keys)
        if not exists:
            return True

        exists.remove(tuple(policy))
        return True

    def __get_policy(self) -> Iterable[Sequence[str]]:
        if self._current_filter is not None:
            return (list(x) for x in self._current_filter)
        else:
            return (list(v2) for v in self._cache.values() for v1 in v.values() for v2 in v1)

    def apply_filter(self, *keys: str) -> None:
        value = in_cache(self._cache, keys)
        self._current_filter = value or set()

    def clear_filter(self) -> None:
        self._current_filter = None


@contextmanager
def fast_policy_filter(policy: FastPolicy, *keys: str) -> Iterator[None]:
    try:
        policy.apply_filter(*keys)
        yield
    finally:
        policy.clear_filter()



# Copyright 2023 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Sequence

from .model import Model
from .policy_fast import FastPolicy


class FastModel(Model):
    _cache_key_order: Sequence[int]

    def __init__(self, cache_key_order: Sequence[int]) -> None:
        super().__init__()
        self._cache_key_order = cache_key_order

    def add_def(self, sec: str, key: str, value: Any) -> None:
        super().add_def(sec, key, value)
        if sec == "p" and key == "p":
            self.model[sec][key].policy = FastPolicy(self._cache_key_order)

    def clear_policy(self) -> None:
        """clears all current policy."""
        super().clear_policy()
        self.model["p"]["p"].policy = FastPolicy(self._cache_key_order)



# Copyright 2023 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Sequence
from unittest import TestCase

import casbin
from casbin import FastPolicy


def get_examples(path):
    examples_path = os.path.split(os.path.realpath(__file__))[0] + "/../examples/"
    return os.path.abspath(examples_path + path)


class TestCaseBase(TestCase):
    def get_enforcer(self, model=None, adapter=None, cache_key_order: Sequence[int] = None):
        return casbin.FastEnforcer(
            model,
            adapter,
            cache_key_order=cache_key_order,
        )


class TestFastEnforcer(TestCaseBase):
    def test_creates_proper_policy(self) -> None:
        e = self.get_enforcer(
            get_examples("basic_model.conf"),
            get_examples("basic_policy.csv"),
            [2, 1],
        )
        self.assertIsInstance(e.model.model["p"]["p"].policy, FastPolicy)

    def test_initializes_model(self) -> None:
        e = self.get_enforcer(
            get_examples("basic_model.conf"),
            get_examples("basic_policy.csv"),
            [2, 1],
        )
        self.assertEqual(
            list(e.model.model["p"]["p"].policy),
            [
                ["alice", "data1", "read"],
                ["bob", "data2", "write"],
            ],
        )

    def test_able_to_clear_policy(self) -> None:
        e = self.get_enforcer(
            get_examples("basic_model.conf"),
            get_examples("basic_policy.csv"),
            [2, 1],
        )

        e.clear_policy()

        self.assertIsInstance(e.model.model["p"]["p"].policy, FastPolicy)
        self.assertEqual(list(e.model.model["p"]["p"].policy), [])

    def test_able_to_enforce_rule(self) -> None:
        e = self.get_enforcer(
            get_examples("basic_model.conf"),
            get_examples("basic_policy.csv"),
            [2, 1],
        )

        self.assertTrue(e.enforce("alice", "data1", "read"))
        self.assertFalse(e.enforce("alice2", "data1", "read"))



# Copyright 2021 The casbin Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from casbin import util, config
from . import Assertion
from .policy import Policy

DEFAULT_DOMAIN = ""
DEFAULT_SEPARATOR = "::"


class Model(Policy):
    section_name_map = {
        "r": "request_definition",
        "p": "policy_definition",
        "g": "role_definition",
        "e": "policy_effect",
        "m": "matchers",
    }

    def _load_assertion(self, cfg, sec, key):
        value = cfg.get(self.section_name_map[sec] + "::" + key)

        return self.add_def(sec, key, value)

    def add_def(self, sec, key, value):
        if value == "":
            return

        ast = Assertion()
        ast.key = key
        ast.value = value

        if "r" == sec or "p" == sec:
            ast.tokens = ast.value.split(",")
            for i, token in enumerate(ast.tokens):
                ast.tokens[i] = key + "_" + token.strip()
        else:
            ast.value = util.remove_comments(util.escape_assertion(ast.value))

        if sec not in self.keys():
            self[sec] = {}

        self[sec][key] = ast

        return True

    def _get_key_suffix(self, i):
        if i == 1:
            return ""

        return str(i)

    def _load_section(self, cfg, sec):
        i = 1
        while True:
            if not self._load_assertion(cfg, sec, sec + self._get_key_suffix(i)):
                break
            else:
                i = i + 1

    def load_model(self, path):
        cfg = config.Config.new_config(path)

        self._load_section(cfg, "r")
        self._load_section(cfg, "p")
        self._load_section(cfg, "e")
        self._load_section(cfg, "m")

        self._load_section(cfg, "g")

    def load_model_from_text(self, text):
        cfg = config.Config.new_config_from_text(text)

        self._load_section(cfg, "r")
        self._load_section(cfg, "p")
        self._load_section(cfg, "e")
        self._load_section(cfg, "m")

        self._load_section(cfg, "g")

    def print_model(self):
        self.logger.info("Model:")
        for k, v in self.items():
            for i, j in v.items():
                self.logger.info("%s.%s: %s", k, i, j.value)

    def sort_policies_by_priority(self):
        for ptype, assertion in self["p"].items():
            for index, token in enumerate(assertion.tokens):
                if token == f"{ptype}_priority":
                    assertion.priority_index = index
                    break

            if assertion.priority_index == -1:
                continue

            assertion.policy = sorted(assertion.policy, key=lambda x: x[assertion.priority_index])

            for i, policy in enumerate(assertion.policy):
                assertion.policy_map[",".join(policy)] = i

        return None

    def sort_policies_by_subject_hierarchy(self):
        if self["e"]["e"].value != "subjectPriority(p_eft) || deny":
            return

        sub_index = 0
        domain_index = -1
        for ptype, assertion in self["p"].items():
            for index, token in enumerate(assertion.tokens):
                if token == "{}_dom".format(ptype):
                    domain_index = index
                    break

            subject_hierarchy_map = self.get_subject_hierarchy_map(self["g"]["g"].policy)

            def compare_policy(policy):
                domain = DEFAULT_DOMAIN
                if domain_index != -1:
                    domain = policy[domain_index]
                name = self.get_name_with_domain(domain, policy[sub_index])
                return subject_hierarchy_map.get(name, 0)

            assertion.policy = sorted(assertion.policy, key=compare_policy)
            for i, policy in enumerate(assertion.policy):
                assertion.policy_map[",".join(policy)] = i

    def get_subject_hierarchy_map(self, policies):
        """
        Get the subject hierarchy from the policy.
        Select the lowest level subject in multiple rounds until all subjects are selected.
        Return the subject hierarchy dictionary, the subject is the key, and the level is the value.
        The level starts from 0 and increases in turn. The smaller the level, the higher the priority.
        """
        # Init unsorted policy, and subject
        unsorted_policy = []
        unsorted_sub = set()
        for policy in policies:
            if len(policy) < 2:
                raise RuntimeError("policy g expect 2 more params")
            domain = DEFAULT_DOMAIN
            if len(policy) != 2:
                domain = policy[2]
            child = self.get_name_with_domain(domain, policy[0])
            parent = self.get_name_with_domain(domain, policy[1])
            unsorted_policy.append([child, parent])
            unsorted_sub.add(child)
            unsorted_sub.add(parent)
        # sort policy,and update sorted_sub_list
        sorted_sub_list = []
        while len(unsorted_policy) > 0:
            # get all parent subject
            parent_sub = {p[1] for p in unsorted_policy if p[1] != ""}
            # remove parent subject from unsorted_sub
            sorted_sub = unsorted_sub - parent_sub
            if not sorted_sub:
                raise RuntimeError("cycle dependency in subject hierarchy.subjects: {}".format(unsorted_sub))
            # update sorted_sub_list
            sorted_sub_list.append(sorted_sub)
            # remove sorted subject, and update unsorted_policy
            unsorted_policy = [p for p in unsorted_policy if p[0] not in sorted_sub]
            # update unsorted_sub
            unsorted_sub = unsorted_sub - sorted_sub
        if len(unsorted_sub) > 0:
            sorted_sub_list.append(unsorted_sub)
        # Tree structure of subject
        return {sub: i for i, subs in enumerate(sorted_sub_list) for sub in subs}

    def get_name_with_domain(self, domain, name):
        return "{}{}{}".format(domain, DEFAULT_SEPARATOR, name)

    def to_text(self):
        s = []

        def write_string(sec):
            for p_type in self[sec]:
                value = self[sec][p_type].value
                s.append("{} = {}\n".format(sec, value.replace("p_", "p.").replace("r_", "r.")))

        s.append("[request_definition]\n")
        write_string("r")
        s.append("[policy_definition]\n")
        write_string("p")
        if "g" in self.keys():
            s.append("[role_definition]\n")
            for p_type in self["g"]:
                s.append("{} = {}\n".format(p_type, self["g"][p_type].value))
        s.append("[policy_effect]\n")
        write_string("e")
        s.append("[matchers]\n")
        write_string("m")

        # remove last \n
        s[-1] = s[-1].strip()

        return "".join(s)

    def get_field_index(self, ptype, field):
        """get_field_index gets the index of the field for a ptype in a policy,
        return -1 if the field does not exist."""
        assertion = self["p"][ptype]
        if field in assertion.field_index_map:
            return assertion.field_index_map[field]

        pattern = f"{ptype}_{field}"
        index = -1
        for i, token in enumerate(assertion.tokens):
            if token == pattern:
                index = i
                break

        if index == -1:
            return index

        assertion.field_index_map[field] = index
        return index
